/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2011-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package
 * @author
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#pragma once

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/system/ImportExportComponent.h>
#include <RobotAPI/libraries/core/FramedPose.h>


#include <RobotAPI/interface/core/RobotState.h>

#include <boost/thread.hpp>

namespace VirtualRobot
{
    class Robot;
    using RobotPtr = boost::shared_ptr<Robot>;
    class RobotNode;
    using RobotNodePtr = boost::shared_ptr<RobotNode>;
}

namespace armarx
{
    // Zugriff muss mutex geschuetzt werden!

    /**
     * @brief The base class for remote objects on the ice server
     *
     */

    class SharedObjectBase :
        virtual public SharedObjectInterface
    {
    public:
        void ref(const Ice::Current& c = Ice::Current()) override;
        void unref(const Ice::Current& c = Ice::Current()) override;
        void destroy(const Ice::Current& c = Ice::Current()) override;
        SharedObjectBase();
    private:
        unsigned int _referenceCount;
        boost::mutex _counterMutex;
    };

    /**
     * @brief The SharedRobotNodeServant class is a remote representation of a Simox VirtualRobot::Robot
     * @ingroup Component-RobotStateComponent
     * @details This class is used only internally by the the RobotStateComponent. Other classes such as the LinkedPose, RemoteRobot,
     * TCPControlUnit and HeadIKUnit classes address this class by the SharedRobotNodeInterface and SharedRobotNodeInterfacePrx generated by
     * Ice.
     */
    class SharedRobotNodeServant :
        virtual public SharedRobotNodeInterface,
        public SharedObjectBase
    {
    public:
        SharedRobotNodeServant(VirtualRobot::RobotNodePtr node /*,const Ice::Current & current = Ice::Current()*/);
        ~SharedRobotNodeServant() override;

        float getJointValue(const Ice::Current& current = Ice::Current()) const override;
        std::string getName(const Ice::Current& current = Ice::Current()) const override;

        PoseBasePtr getLocalTransformation(const Ice::Current& current = Ice::Current()) const override;
        FramedPoseBasePtr getGlobalPose(const Ice::Current& current = Ice::Current()) const override;
        FramedPoseBasePtr getPoseInRootFrame(const Ice::Current& current = Ice::Current()) const override;

        JointType getType(const Ice::Current& current = Ice::Current()) const override;
        Vector3BasePtr getJointTranslationDirection(const Ice::Current& current = Ice::Current()) const override;
        Vector3BasePtr getJointRotationAxis(const Ice::Current& current = Ice::Current()) const override;

        bool hasChild(const std::string& name, bool recursive, const Ice::Current& current = Ice::Current()) const override;
        std::string getParent(const Ice::Current& current = Ice::Current()) const override;
        NameList getChildren(const Ice::Current& current = Ice::Current()) const override;
        NameList getAllParents(const std::string& name, const Ice::Current& current = Ice::Current()) const override;

        float getJointValueOffest(const Ice::Current& current = Ice::Current()) const override;
        float getJointLimitHigh(const Ice::Current& current = Ice::Current()) const override;
        float getJointLimitLow(const Ice::Current& current = Ice::Current()) const override;

        Vector3BasePtr getCoM(const Ice::Current& current = Ice::Current()) const override;
        Ice::FloatSeq getInertia(const Ice::Current& current = Ice::Current()) const override;
        float getMass(const Ice::Current& current = Ice::Current()) const override;



    protected:
        VirtualRobot::RobotNodePtr _node;
    };

    /**
     * @ingroup Component-RobotStateComponent
     * @brief The SharedRobotNodeServant class is a remote represenation of a Simox VirtualRobot::Robot
     *
     * @details This class is used only internally by the the RobotStateComponent. The RemoteRobot class SharedRobotInterface and SharedRobotInterfacePrx
     * classes generated by Ice.
     */

    class   SharedRobotServant :
        public virtual SharedRobotInterface,
        public SharedObjectBase
    {
    public:
        SharedRobotServant(VirtualRobot::RobotPtr robot, RobotStateComponentInterfacePrx robotStateComponent, RobotStateListenerInterfacePrx robotStateListenerPrx);
        ~SharedRobotServant() override;
        void setRobotStateComponent(RobotStateComponentInterfacePrx robotStateComponent);
        SharedRobotNodeInterfacePrx getRobotNode(const std::string& name, const Ice::Current& current = Ice::Current()) override;
        SharedRobotNodeInterfacePrx getRootNode(const Ice::Current& current = Ice::Current()) override;
        bool hasRobotNode(const std::string& name, const Ice::Current& current = Ice::Current()) override;

        NameList getRobotNodes(const Ice::Current& current = Ice::Current()) override;
        RobotNodeSetInfoPtr getRobotNodeSet(const std::string& name, const Ice::Current& current = Ice::Current()) override;
        NameList getRobotNodeSets(const Ice::Current& current = Ice::Current()) override;
        bool hasRobotNodeSet(const std::string& name, const Ice::Current& current = Ice::Current()) override;

        std::string getName(const Ice::Current& current = Ice::Current()) override;
        std::string getType(const Ice::Current& current = Ice::Current()) override;

        PoseBasePtr getGlobalPose(const Ice::Current& current = Ice::Current()) override;
        NameValueMap getConfig(const Ice::Current& current = Ice::Current()) override;
        NameValueMap getConfigAndPose(PoseBasePtr& globalPose, const Ice::Current& current = Ice::Current()) override;
        void setGlobalPose(const PoseBasePtr& pose, const Ice::Current& current = Ice::Current()) override;

        float getScaling(const Ice::Current& = Ice::Current()) override;

        VirtualRobot::RobotPtr getRobot() const;

        void setTimestamp(const IceUtil::Time& updateTime);
        TimestampBasePtr getTimestamp(const Ice::Current& = Ice::Current()) const override;
        RobotStateComponentInterfacePrx getRobotStateComponent(const Ice::Current&) const override;
    protected:
        VirtualRobot::RobotPtr _robot;
        boost::recursive_mutex m;
        std::map<std::string, SharedRobotNodeInterfacePrx> _cachedNodes;
        IceUtil::Time updateTimestamp;
        RobotStateComponentInterfacePrx robotStateComponent;
        RobotStateListenerInterfacePrx robotStateListenerPrx;

    };

    using SharedRobotServantPtr = IceInternal::Handle<SharedRobotServant>;
}

