/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    RobotAPI
* @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
* @date       2017
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "BasicControllers.h"

#include <boost/algorithm/clamp.hpp>

#include <ArmarXCore/core/util/algorithm.h>
#include <ArmarXCore/core/logging/Logging.h>

namespace armarx
{
    float velocityControlWithAccelerationBounds(
        float dt, float maxDt,
        const float currentV, float targetV, float maxV,
        float acceleration, float deceleration,
        float directSetVLimit
    )
    {
        dt = std::min(std::abs(dt), std::abs(maxDt));
        maxV = std::abs(maxV);
        acceleration = std::abs(acceleration);
        deceleration = std::abs(deceleration);
        targetV = boost::algorithm::clamp(targetV, -maxV, maxV);

        //we can have 3 cases:
        // 1. we directly set v and ignore acc/dec (if |curr - target| <= limit)
        // 2. we need to accelerate                (if curr and target v have same sign and |curr| < |target|)
        // 3. we need to decelerate               (other cases)

        //handle case 1
        const float curverror = targetV - currentV;
        if (std::abs(curverror) < directSetVLimit)
        {
            return targetV;
        }

        //handle case 2 + 3
        const bool accelerate = sign(targetV) == sign(currentV) &&     // v in same direction
                                std::abs(targetV) > std::abs(currentV); // currently to slow

        const float usedacc = accelerate ? acceleration : -deceleration;
        const float maxDeltaV = std::abs(usedacc * dt);
        if (maxDeltaV >= std::abs(curverror))
        {
            //we change our v too much with the used acceleration
            //but we could reach our target with a lower acceleration ->set the target

            return targetV;
        }
        const float deltaVel = boost::algorithm::clamp(sign(currentV) * usedacc * dt, -maxDeltaV, maxDeltaV);
        const float nextV = currentV + deltaVel;
        return nextV;
    }

    ////////////////////////////
    //wip?
    float velocityControlWithAccelerationAndPositionBounds(float dt, float maxDt,
            float currentV, float targetV, float maxV,
            float acceleration, float deceleration,
            float directSetVLimit,
            float currentPosition,
            float positionLimitLoSoft, float positionLimitHiSoft,
            float positionLimitLoHard, float positionLimitHiHard)
    {
        if (currentPosition <= positionLimitLoHard || currentPosition >= positionLimitHiHard)
        {
            return std::nanf("1");
        }

        float softLimitViolation = 0;
        if (currentPosition <= positionLimitLoSoft)
        {
            softLimitViolation = -1;
        }
        if (currentPosition >= positionLimitHiSoft)
        {
            softLimitViolation = 1;
        }

        const float upperDt = std::max(std::abs(dt), std::abs(maxDt));
        dt = std::min(std::abs(dt), std::abs(maxDt));
        maxV = std::abs(maxV);

        //we can have 4 cases:
        // 1. we need to decelerate now or we will violate the position limits (maybe we still will violate them, e.g. if we violated them initially)
        // 2. we directly set v and ignore acc/dec (if |curr - target| <= limit)
        // 3. we need to accelerate                (if curr and target v have same sign and |curr| < |target|)
        // 4. we need to decelerate               (other cases)
        float nextv;
        //handle case 1
        const float vsquared = currentV * currentV;
        const float brakingDist = sign(currentV) * vsquared / 2.f / std::abs(deceleration); //the braking distance points in the direction of the velocity


        const float posIfBrakingNow = currentPosition + brakingDist;
        if (posIfBrakingNow <= positionLimitLoSoft || posIfBrakingNow >= positionLimitHiSoft)
        {
            //case 1. -> brake now! (we try to have v=0 at the limit)
            const auto limit = posIfBrakingNow <= positionLimitLoSoft ? positionLimitLoSoft : positionLimitHiSoft;
            const float wayToGo = limit - currentPosition;

            //decelerate!
            // s = v²/(2a)  <=>  a = v²/(2s)
            const float dec = std::abs(vsquared / 2.f / wayToGo);
            const float vel = currentV - sign(currentV) * dec * upperDt;
            nextv = boost::algorithm::clamp(vel, -maxV, maxV);
            if (sign(currentV) != sign(nextv))
            {
                //stop now
                nextv = 0;
            }
        }
        else
        {
            //handle 2-3
            nextv = velocityControlWithAccelerationBounds(dt, maxDt, currentV, targetV, maxV, acceleration, deceleration, directSetVLimit);
        }
        if (softLimitViolation == sign(nextv))
        {
            //the area between soft and hard limits is sticky
            //the controller can only move out of it (not further in)
            return 0;
        }

        //the next velocity will not violate the pos bounds harder than they are already
        return nextv;
    }


    float positionThroughVelocityControlWithAccelerationBounds(
        float dt, float maxDt,
        float currentV, float maxV,
        float acceleration, float deceleration,
        float currentPosition, float targetPosition,
        float p
    )
    {
        dt = std::min(std::abs(dt), std::abs(maxDt));
        maxV = std::abs(maxV);
        acceleration = std::abs(acceleration);
        deceleration = std::abs(deceleration);
        const float signV = sign(currentV);
        //we can have 3 cases:
        // 1. we use a p controller and ignore acc/dec (if velocity target from p controller < velocity target from ramp controller)
        // 2. we need to accelerate (or hold vel)      (if e = (targetPosition - currentPosition)
        //                                                 the brakingDistance have the same sign and brakingDistance < e
        //                                                 and currentVel <= maxV)
        // 3. we need to decelerate                   (other cases)

        //handle case 1
        const float positionError = targetPosition - currentPosition;
        float newTargetVelPController = positionError * p;

        //handle case 2-3
        const float brakingDistance = signV * currentV * currentV / 2.f / deceleration; //the braking distance points in the direction of the velocity
        const float posIfBrakingNow = currentPosition + brakingDistance;
        const float posErrorIfBrakingNow = targetPosition - posIfBrakingNow;

        const bool decelerate =
            std::abs(currentV) > maxV || // we need to slow down (to stay in [-maxV,maxV]
            //            std::abs(posIfBrakingNow - targetPosition) <= pControlPosErrorLimit || // we want to hit the target
            sign(posErrorIfBrakingNow) != signV;  // we are moving away from the target
        const float usedacc = decelerate ?  -deceleration : acceleration;
        const float maxDeltaV = std::abs(usedacc * dt);
        const float deltaVel = boost::algorithm::clamp(signV * usedacc * dt, -maxDeltaV, maxDeltaV);

        float newTargetVel = boost::algorithm::clamp(currentV + deltaVel, -maxV, maxV);
        bool usePID = std::abs(newTargetVelPController) < std::abs(newTargetVel);
        //        ARMARX_INFO << deactivateSpam(0.01) << VAROUT(usePID) << VAROUT(dt) << VAROUT(decelerate) << VAROUT(usedacc) << VAROUT(maxDeltaV) << VAROUT(deltaVel) << VAROUT(newTargetVel) << VAROUT(newTargetVelPController)
        //                    << VAROUT(currentPosition) << VAROUT(targetPosition);
        if (usePID)
        {
            return newTargetVelPController;
        }
        else
        {
            return newTargetVel;
        }
    }

    float positionThroughVelocityControlWithAccelerationAndPositionBounds(
        float dt, float maxDt,
        float currentV, float maxV,
        float acceleration, float deceleration,
        float currentPosition, float targetPosition,
        float p,
        float positionLimitLo, float positionLimitHi

    )
    {
        dt = std::min(std::abs(dt), std::abs(maxDt));
        maxV = std::abs(maxV);
        acceleration = std::abs(acceleration);
        deceleration = std::abs(deceleration);
        const float signV = sign(currentV);
        //we can have 4 cases:
        // 1. we need to decelerate now or we will violate the position limits (maybe we still will violate them)
        // 2. we use a p controller and ignore acc/dec (if |currentPosition - targetPosition| <= pControlPosErrorLimit AND |currentPosition - targetPosition|*p < maxV)
        // 4. we need to accelerate (or hold vel)      (if e = (targetPosition - currentPosition)
        //                                                 and the brakingDistance have the same sign and brakingDistance < e
        //                                                 and currentVel <= maxV)
        // 5. we need to decelerate                   (other cases)

        //handle case 1
        const float vsquared = currentV * currentV;
        const float brakingDistance = signV * vsquared / 2.f / deceleration; //the braking distance points in the direction of the velocity
        const float posIfBrakingNow = currentPosition + brakingDistance;
        if (posIfBrakingNow <= positionLimitLo || posIfBrakingNow >= positionLimitHi)
        {
            //case 1. -> brake now! (we try to have v=0 at the limit)
            const auto limit = brakingDistance > 0 ? positionLimitHi : positionLimitLo;
            const float wayToGo = std::abs(limit - currentPosition);
            // s = v²/(2a)  <=>  a = v²/(2s)
            const float dec = std::abs(vsquared / 2.f / wayToGo);
            const float vel = currentV - signV * dec * dt;
            return vel;
        }

        //handle case 2-3
        return positionThroughVelocityControlWithAccelerationBounds(
                   dt, maxDt,
                   currentV, maxV,
                   acceleration, deceleration,
                   currentPosition, boost::algorithm::clamp(targetPosition, positionLimitLo, positionLimitHi),
                   p
               );
    }

    ////////////////////////////
    //wip


    float positionThroughVelocityControlWithAccelerationBoundsAndPeriodicPosition(
        float dt, float maxDt,
        float currentV, float maxV,
        float acceleration, float deceleration,
        float currentPosition, float targetPosition,
        float pControlPosErrorLimit, float p,
        float& direction,
        float positionPeriodLo, float positionPeriodHi)
    {
        currentPosition = periodicClamp(currentPosition, positionPeriodLo, positionPeriodHi);
        targetPosition  = periodicClamp(targetPosition, positionPeriodLo, positionPeriodHi);

        const float brakingDist = brakingDistance(currentV, deceleration);
        const float posIfBrakingNow = currentPosition + brakingDist;
        const float posIfBrakingNowError = targetPosition - posIfBrakingNow;
        const float posError = targetPosition - currentPosition;
        if (
            std::abs(posIfBrakingNowError) <= pControlPosErrorLimit ||
            std::abs(posError) <= pControlPosErrorLimit
        )
        {
            //this allows slight overshooting (in the limits of the p controller)
            return positionThroughVelocityControlWithAccelerationBounds(
                       dt, maxDt,
                       currentV, maxV,
                       acceleration, deceleration,
                       currentPosition, targetPosition,
                       p
                   );
        }

        //we transform the problem with periodic bounds into
        //a problem with position bounds
        const float positionPeriodLength  = std::abs(positionPeriodHi - positionPeriodLo);

        //we shift the positions such that 0 == target
        currentPosition = periodicClamp(currentPosition - targetPosition,  0.f, positionPeriodLength);
        //how many times will we go ovet the target if we simply bake now?
        const float overshoot = std::trunc((currentPosition + brakingDist) / positionPeriodLength);

        if (true || direction == 0)
        {
            //determine the direction to go (1 == pos vel, -1 == neg vel)
            direction = (periodicClamp(currentPosition + brakingDist, 0.f, positionPeriodLength) >= positionPeriodLength / 2) ? 1 : -1;
        }
        //shift the target away from 0
        targetPosition = (overshoot - std::min(0.f, -direction)) * positionPeriodLength; // - direction * pControlPosErrorLimit;

        //move
        return positionThroughVelocityControlWithAccelerationBounds(
                   dt, maxDt,
                   currentV, maxV,
                   acceleration, deceleration,
                   currentPosition, targetPosition,
                   p
               );
    }

    bool VelocityControllerWithAccelerationBounds::validParameters() const
    {
        return  maxV > 0 &&
                acceleration > 0 &&
                deceleration > 0 &&
                targetV <=  maxV &&
                targetV >= -maxV;
    }

    float VelocityControllerWithAccelerationBounds::run() const
    {
        const float useddt = std::min(std::abs(dt), std::abs(maxDt));

        //we can have 3 cases:
        // 1. we directly set v and ignore acc/dec (if |curr - target| <= limit)
        // 2. we need to accelerate                (if curr and target v have same sign and |curr| < |target|)
        // 3. we need to decelerate               (other cases)
        //handle case 1
        const float curverror = targetV - currentV;
        if (std::abs(curverror) < directSetVLimit)
        {
            return targetV;
        }

        //handle case 2 + 3
        const bool accelerate = sign(targetV) == sign(currentV) &&     // v in same direction
                                std::abs(targetV) > std::abs(currentV); // currently to slow

        const float usedacc = accelerate ? acceleration : -deceleration;
        const float maxDeltaV = std::abs(usedacc * useddt);
        if (maxDeltaV >= std::abs(curverror))
        {
            //we change our v too much with the used acceleration
            //but we could reach our target with a lower acceleration ->set the target

            return targetV;
        }
        const float deltaVel = boost::algorithm::clamp(sign(currentV) * usedacc * useddt, -maxDeltaV, maxDeltaV);
        const float nextV = currentV + deltaVel;
        return nextV;
    }

    float VelocityControllerWithAccelerationBounds::estimateTime() const
    {
        const float curverror = std::abs(targetV - currentV);
        if (curverror < directSetVLimit)
        {
            return maxDt;
        }
        return curverror / ((targetV > currentV) ? acceleration : deceleration);
    }

    bool VelocityControllerWithAccelerationAndPositionBounds::validParameters() const
    {
        return  VelocityControllerWithAccelerationBounds::validParameters() &&
                //                positionLimitLoHard < positionLimitLoSoft &&
                positionLimitLoSoft < positionLimitHiSoft;
        //                positionLimitHiSoft < positionLimitHiHard;
    }

    float VelocityControllerWithAccelerationAndPositionBounds::run() const
    {
        //        if (currentPosition <= positionLimitLoHard || currentPosition >= positionLimitHiHard)
        //        {
        //            ARMARX_INFO << deactivateSpam(1) << "Hard limit violation. " << VAROUT(currentPosition) << VAROUT(positionLimitLoHard) << VAROUT(positionLimitHiHard);
        //            return std::nanf("1");
        //        }

        float softLimitViolation = 0;
        if (currentPosition <= positionLimitLoSoft)
        {
            softLimitViolation = -1;
        }
        if (currentPosition >= positionLimitHiSoft)
        {
            softLimitViolation = 1;
        }

        const float upperDt = std::max(std::abs(dt), std::abs(maxDt));

        //we can have 4 cases:
        // 1. we need to decelerate now or we will violate the position limits (maybe we still will violate them, e.g. if we violated them initially)
        // 2. we directly set v and ignore acc/dec (if |curr - target| <= limit)
        // 3. we need to accelerate                (if curr and target v have same sign and |curr| < |target|)
        // 4. we need to decelerate               (other cases)
        float nextv;
        //handle case 1
        const float vsquared = currentV * currentV;
        const float brakingDist = sign(currentV) * vsquared / 2.f / std::abs(deceleration); //the braking distance points in the direction of the velocity

        const float posIfBrakingNow = currentPosition + brakingDist;
        //        ARMARX_INFO << deactivateSpam(1) << VAROUT(currentPosition) << VAROUT(positionLimitLoSoft) << VAROUT(positionLimitHiSoft) << VAROUT(posIfBrakingNow);
        if ((posIfBrakingNow <= positionLimitLoSoft && currentV < 0)  || (posIfBrakingNow >= positionLimitHiSoft && currentV > 0))
        {
            //case 1. -> brake now! (we try to have v=0 at the limit)
            const auto limit = posIfBrakingNow <= positionLimitLoSoft ? positionLimitLoSoft : positionLimitHiSoft;
            const float wayToGo = limit - currentPosition;

            //decelerate!
            // s = v²/(2a)  <=>  a = v²/(2s)
            const float dec = std::abs(vsquared / 2.f / wayToGo);
            const float vel = currentV - sign(currentV) * dec * upperDt;
            nextv = boost::algorithm::clamp(vel, -maxV, maxV);
            //            ARMARX_INFO << deactivateSpam(1) << "clamped new Vel: " << VAROUT(nextv);
            if (sign(currentV) != sign(nextv))
            {
                //                ARMARX_INFO << deactivateSpam(1) << "wrong sign: stopping"; //stop now
                nextv = 0;
            }
        }
        else
        {
            //handle 2-3
            nextv = VelocityControllerWithAccelerationBounds::run();
            //            ARMARX_INFO << deactivateSpam(1) << "Desired target Vel: " << targetV << " " << VAROUT(nextv);

        }
        if (softLimitViolation == sign(nextv) && nextv != 0)
        {
            //the area between soft and hard limits is sticky
            //the controller can only move out of it (not further in)
            //            ARMARX_DEBUG << deactivateSpam(1) << "Soft limit violation. " << softLimitViolation << VAROUT(nextv);
            return 0;
        }

        //the next velocity will not violate the pos bounds harder than they are already
        return nextv;
    }

    PositionThroughVelocityControllerWithAccelerationBounds::PositionThroughVelocityControllerWithAccelerationBounds()
    {
#ifdef DEBUG_POS_CTRL
        buffer = boost::circular_buffer<HelpStruct>(20);
#endif
    }

    bool PositionThroughVelocityControllerWithAccelerationBounds::validParameters() const
    {
        return  maxV > 0 &&
                acceleration > 0 &&
                deceleration > 0 &&
                //                pControlPosErrorLimit > 0 &&
                //                pControlVelLimit > 0 &&
                p > 0;
    }

    float PositionThroughVelocityControllerWithAccelerationBounds::run() const
    {
        const float useddt = std::min(std::abs(dt), std::abs(maxDt));
        const float signV = sign(currentV);
        //we can have 3 cases:
        // 1. we use a p controller and ignore acc/dec (if |currentPosition - targetPosition| <= pControlPosErrorLimit AND |currentV| < pControlVelLimit)
        // 2. we need to accelerate (or hold vel)      (if e = (targetPosition - currentPosition)
        //                                                 the brakingDistance have the same sign and brakingDistance < e
        //                                                 and currentVel <= maxV)
        // 3. we need to decelerate                   (other cases)

        //handle case 1
        const float positionError = targetPosition - currentPosition;
        float newTargetVelPController = positionError * p;

        //handle case 2-3
        const float brakingDistance = signV * currentV * currentV / 2.f / deceleration; //the braking distance points in the direction of the velocity
        const float posIfBrakingNow = currentPosition + brakingDistance;
        const float posErrorIfBrakingNow = targetPosition - posIfBrakingNow;
        const bool hardBrakingNeeded = std::abs(brakingDistance) > std::abs(positionError);
        const float usedDeceleration = hardBrakingNeeded ?
                                       std::abs(currentV * currentV / 2.f / math::MathUtils::LimitTo(positionError, 0.0001)) :
                                       deceleration;
        const bool decelerate =
            std::abs(currentV) > maxV || // we need to slow down (to stay in [-maxV,maxV]
            hardBrakingNeeded ||
            sign(posErrorIfBrakingNow) != signV;  // we are moving away from the target
        const float usedacc = decelerate ?  -usedDeceleration : acceleration;
        const float deltaVel = signV * usedacc * useddt;
        float newTargetVelRampCtrl = boost::algorithm::clamp(currentV + deltaVel, -maxV, maxV);
        bool usePID = std::abs(newTargetVelPController) < std::abs(newTargetVelRampCtrl)
                      || std::abs(newTargetVelPController) < pControlVelLimit
                      || std::abs(positionError) < pControlPosErrorLimit;
        float finalTargetVel = usePID ? newTargetVelPController : newTargetVelRampCtrl;
        if (std::abs(positionError) < accuracy)
        {
            finalTargetVel = 0.0f;    // if close to target set zero velocity to avoid oscillating around target
        }

#ifdef DEBUG_POS_CTRL
        buffer.push_back({currentPosition, newTargetVelPController, newTargetVelRampCtrl, currentV, positionError, IceUtil::Time::now().toMicroSeconds()});

        //        if (PIDModeActive != usePID)
        if (buffer.size() > 0 && sign(positionError) * sign(buffer[buffer.size() - 2].currentError) < 0 && eventHappeningCounter < 0)
        {
            eventHappeningCounter = 10;
            ARMARX_IMPORTANT << "HIGH VELOCITY  DETECTED";
        }
        if (eventHappeningCounter == 0)
        {
            ARMARX_IMPORTANT << "BUFFER CONTENT";
            for (auto& elem : buffer)
            {
                ARMARX_INFO << VAROUT(elem.currentPosition) << VAROUT(elem.targetVelocityPID) << VAROUT(elem.targetVelocityRAMP) << VAROUT(elem.currentV) << VAROUT(elem.currentError) << VAROUT(elem.timestamp);
            }
            ARMARX_IMPORTANT << VAROUT(newTargetVelPController) << VAROUT(newTargetVelRampCtrl) << VAROUT(currentPosition) << VAROUT(usePID) << VAROUT(positionError);

        }
        if (eventHappeningCounter >= 0)
        {
            eventHappeningCounter--;
        }

        ARMARX_INFO << deactivateSpam(5) << VAROUT(newTargetVelPController) << VAROUT(newTargetVelRampCtrl) << VAROUT(currentPosition) << VAROUT(usePID) << VAROUT(positionError) << VAROUT(targetPosition) << VAROUT(currentV) << VAROUT(deltaVel) << VAROUT(useddt) << VAROUT(usedacc);
        PIDModeActive = usePID;
#endif

        return finalTargetVel;
    }

    float PositionThroughVelocityControllerWithAccelerationBounds::estimateTime() const
    {
        //            const float brakingDistance = sign(currentV) * currentV * currentV / 2.f / deceleration;
        const float posError = targetPosition - currentPosition;
        const float posIfBrakingNow = currentPosition + brakingDistance(currentV, deceleration);
        //we are near the goal if we break now
        //        if (std::abs(posIfBrakingNow - targetPosition) <= pControlPosErrorLimit  && currentV <= pControlVelLimit)
        //        {
        //            return std::sqrt(2 * std::abs(brakingDistance(currentV, deceleration)) / deceleration);
        //        }
        const float posErrorIfBrakingNow = targetPosition - posIfBrakingNow;

        float t = 0;
        float curVel = currentV;
        float curPos = currentPosition;
        auto curPosEr = [&] {return targetPosition - curPos;};
        //            auto curBrake = [&] {return std::abs(brakingDistance(curVel, deceleration));};

        //check for overshooting of target
        if (sign(posError) != sign(posErrorIfBrakingNow))
        {
            t = currentV / deceleration;
            curVel = 0;
            curPos = posIfBrakingNow;
        }
        //        //check for to high v
        //        if (std::abs(curVel) > maxV)
        //        {
        //            const float dv = curVel - maxV;
        //            float dt = dv / deceleration;
        //            t += dt;
        //            //            curPos += 0.5f / deceleration * dv * dv * sign(curVel);
        //            curPos += 0.5 * deceleration * dt * dt + curVel * dt * sign(curVel);
        //            curVel = maxV;
        //        }
        //curBrake()<=curPosEr() && curVel <= maxV
        //        auto tr = trapeze(curVel, acceleration, maxV, deceleration, 0, curPosEr());
        auto tr = trapeze(curVel, std::abs(curVel) > maxV ? deceleration : acceleration, maxV, deceleration, 0, curPosEr());

        return t + tr.at(0).dt + tr.at(1).dt + tr.at(2).dt;


    }

    bool PositionThroughVelocityControllerWithAccelerationAndPositionBounds::validParameters() const
    {
        return  PositionThroughVelocityControllerWithAccelerationBounds::validParameters() &&
                ! std::isnan(positionLimitLo) &&
                ! std::isnan(positionLimitHi) &&
                positionLimitLo < positionLimitHi &&
                targetPosition <= positionLimitHi &&
                positionLimitLo <= targetPosition;
    }

    float PositionThroughVelocityControllerWithAccelerationAndPositionBounds::run() const
    {
        const float useddt = std::min(std::abs(dt), std::abs(maxDt));
        const float signV = sign(currentV);
        //we can have 4 cases:
        // 1. we need to decelerate now or we will violate the position limits (maybe we still will violate them)
        // 2. we use a p controller and ignore acc/dec (if |currentPosition - targetPosition| <= pControlPosErrorLimit AND |currentPosition - targetPosition|*p < maxV)
        // 4. we need to accelerate (or hold vel)      (if e = (targetPosition - currentPosition)
        //                                                 and the brakingDistance have the same sign and brakingDistance < e
        //                                                 and currentVel <= maxV)
        // 5. we need to decelerate                   (other cases)

        //handle case 1
        const float vsquared = currentV * currentV;
        const float brakingDistance = signV * vsquared / 2.f / deceleration; //the braking distance points in the direction of the velocity
        const float posIfBrakingNow = currentPosition + brakingDistance;
        float direction = targetPosition < currentPosition ? -1 : 1;
        //        ARMARX_INFO << deactivateSpam(0.5) << VAROUT(direction) << VAROUT(brakingDistance) << VAROUT(currentPosition) << VAROUT(targetPosition);
        if ((posIfBrakingNow <= positionLimitLo && direction < 1) ||
            (posIfBrakingNow >= positionLimitHi && direction > 1))
        {
            //case 1. -> brake now! (we try to have v=0 at the limit)
            const auto limit = brakingDistance > 0 ? positionLimitHi : positionLimitLo;
            const float wayToGo = std::abs(limit - currentPosition);
            // s = v²/(2a)  <=>  a = v²/(2s)
            const float dec = std::abs(vsquared / 2.f / wayToGo);
            const float vel = currentV - signV * dec * useddt;

            return vel;
        }

        //handle case 2-3
        return PositionThroughVelocityControllerWithAccelerationBounds::run();
    }

    std::array<deltas, 3> trapeze(float v0, float acc, float vMax, float dec, float vt, float dx)
    {
        acc  = std::abs(acc);
        vMax = std::abs(vMax);
        dec  = std::abs(dec);
        auto calc = [&](float vmax)
        {
            const deltas dacc = accelerateToVelocity(v0, acc, vmax);
            const deltas ddec = accelerateToVelocity(vmax, dec, vt);
            const float dxMax = dacc.dx + ddec.dx;
            if (sign(dxMax) == sign(dx))
            {
                if (std::abs(dxMax) <= dx)
                {
                    deltas mid;
                    mid.dx = dxMax - dx;
                    mid.dv = vMax;
                    mid.dt = std::abs(mid.dx / mid.dv);
                    return std::make_pair(true, std::array<deltas, 3> {dacc, mid, ddec});
                }
                //we need a lower vMax (vx)
                //in all following formulas # elem {0,t}
                //a0=acc , at = dec
                //vx=v0+dv0 = vt + dvt -> dv# = vx-v#
                //dx=dx0+dxt
                //dx#=dv#^2/2/a#+dv#/a#*v#
                //dx#=(vx-v#)^2/2/a#+(vx-v#)/a#*v#
                //dx#=vx^2*(1/2/a#) + vx*(v#/2 - v#/a#) + (v#^2/2/a# - v#^2/2)
                //dx=vx^2*(1/2/a0+1/2/at) + vx*(v0/2 - v0/a0 + vt/2 - vt/at) + (v0^2/2/a0 - v0^2/2 +vt^2/2/at - vt^2/2)
                //0 =vx^2*(1/2/a0+1/2/at) + vx*(v0/2 - v0/a0 + vt/2 - vt/at) + (v0^2/2/a0 - v0^2/2 +vt^2/2/at - vt^2/2 -dx)
                //div = (1/2/a0             + 1/2/at                )
                //p   = (v0/2 - v0/a0       + vt/2 - vt/at          )/div
                //q   = (v0^2/2/a0 - v0^2/2 + vt^2/2/at - vt^2/2 -dx)/div
                // -> vx1/2 = pq(p,q)
                const float a0 = acc;
                const float at = dec;
                const float div = (1.f / 2.f / a0               + 1.f / 2.f / at);
                const float p   = (v0 / 2.f - v0 / a0           + vt / 2.f - vt / at) / div;
                const float q   = (v0 * v0 / 2.f * (1.f / a0 - 1.f) + vt * vt / 2.f * (1.f / at - 1.f) - dx) / div;
                const auto vxs = pq(p, q);
                //one or both of the results may be invalid
                bool vx1Valid = std::isfinite(vxs.first) && std::abs(vxs.first) <= std::abs(vmax);
                bool vx2Valid = std::isfinite(vxs.second) && std::abs(vxs.second) <= std::abs(vmax);
                switch (vx1Valid + vx2Valid)
                {
                    case 0:
                        return std::make_pair(false, std::array<deltas, 3>());
                    case 1:
                    {
                        float vx = vx1Valid ? vxs.first : vxs.second;
                        const deltas daccvx = accelerateToVelocity(v0, acc, vx);
                        const deltas ddecvx = accelerateToVelocity(vx, dec, vt);
                        deltas mid;
                        mid.dx = 0;
                        mid.dv = vx;
                        mid.dt = 0;
                        return std::make_pair(true, std::array<deltas, 3> {daccvx, mid, ddecvx});
                    }
                    case 2:
                    {
                        const deltas daccvx1 = accelerateToVelocity(v0, acc, vxs.first);
                        const deltas ddecvx1 = accelerateToVelocity(vxs.first, dec, vt);
                        const deltas daccvx2 = accelerateToVelocity(v0, acc, vxs.second);
                        const deltas ddecvx2 = accelerateToVelocity(vxs.second, dec, vt);
                        deltas mid;
                        mid.dx = 0;
                        mid.dt = 0;
                        if (daccvx1.dt + ddecvx1.dt < daccvx2.dt + ddecvx2.dt)
                        {
                            mid.dv = vxs.first;
                            return std::make_pair(true, std::array<deltas, 3> {daccvx1, mid, ddecvx1});
                        }
                        mid.dv = vxs.second;
                        return std::make_pair(true, std::array<deltas, 3> {daccvx2, mid, ddecvx2});
                    }
                    default:
                        throw std::logic_error {"unreachable code (bool + bool neither 0,1,2)"};
                }
            }
            return std::make_pair(false, std::array<deltas, 3>());
        };

        const auto plusVMax = calc(vMax);
        const auto negVMax  = calc(-vMax);
        switch (plusVMax.first + negVMax.first)
        {
            case 0:
                throw std::invalid_argument("could not find a trapez to reach the goal");
            case 1:
                return plusVMax.first ? plusVMax.second : negVMax.second;
            case 2:
            {
                const float dt1 = plusVMax.second.at(0).dt + plusVMax.second.at(1).dt + plusVMax.second.at(2).dt;
                const float dt2 = negVMax .second.at(0).dt + negVMax .second.at(1).dt + negVMax .second.at(2).dt;
                return dt1 < dt2 ? plusVMax.second : negVMax.second;
            }
            default:
                throw std::logic_error {"unreachable code (bool + bool neither 0,1,2)"};
        }
    }





    float PositionThroughVelocityControllerWithAccelerationBoundsAndPeriodicPosition::run() const
    {
        //run should be static, but we need to adapt the parabs for the subcontroler -> have the subcontroler as value;
        PositionThroughVelocityControllerWithAccelerationBounds sub = *this; // slice parent class from this
        sub.currentPosition = periodicClamp(currentPosition, positionPeriodLo, positionPeriodHi);
        //        ARMARX_INFO << deactivateSpam(1) << VAROUT(currentPosition) << VAROUT(sub.currentPosition) << VAROUT(positionPeriodLo) << VAROUT(positionPeriodHi);
        sub.targetPosition  = periodicClamp(targetPosition, positionPeriodLo, positionPeriodHi);

        const float brakingDist = brakingDistance(currentV, deceleration);
        //        const float posIfBrakingNow = currentPosition + brakingDist;
        //        const float posIfBrakingNowError = targetPosition - posIfBrakingNow;
        //        const float posError = targetPosition - currentPosition;
        //        if (
        //            std::abs(posIfBrakingNowError) <= pControlPosErrorLimit ||
        //            std::abs(posError) <= pControlPosErrorLimit
        //        )
        //        {
        //            //this allows slight overshooting (in the limits of the p controller)
        //            return PositionThroughVelocityControllerWithAccelerationBounds::run();
        //        }

        //we transform the problem with periodic bounds into
        //a problem with position bounds
        const float positionPeriodLength  = std::abs(positionPeriodHi - positionPeriodLo);

        //we shift the positions such that 0 == target
        sub.currentPosition = periodicClamp(currentPosition - targetPosition,  -positionPeriodLength * 0.5f, positionPeriodLength * 0.5f);
        //how many times will we go over the target if we simply brake now?
        const float overshoot = std::trunc((currentPosition + brakingDist) / positionPeriodLength);

        ///TODO check if there may be a case requiring the direction
        float direction = 0;
        if (direction == 0)
        {
            //determine the direction to go (1 == pos vel, -1 == neg vel)
            direction = (periodicClamp(currentPosition + brakingDist, -positionPeriodLength * 0.5f, positionPeriodLength * 0.5f) <= 0) ? 1 : -1;
        }
        //shift the target away from 0
        sub.targetPosition = (overshoot * -direction) * positionPeriodLength; // - direction * pControlPosErrorLimit;

        //move
        return sub.run();
    }





}
