/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ActuatorVelocityTarget
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include "ControlTargetBase.h"
#include <ArmarXCore/observers/variant/Variant.h>

namespace armarx
{
#define make_ControlTarget1DoFActuator(name,varname,cmode)                                                          \
    class name: public ControlTargetBase                                                                            \
    {                                                                                                               \
    public:                                                                                                         \
        float varname = ControllerConstants::ValueNotSetNaN;                                                        \
        name() = default;                                                                                           \
        name(const name&) = default;                                                                                \
        name(name&&) = default;                                                                                     \
        name(float val) : varname{val} {}                                                                           \
        name& operator=(const name&) = default;                                                                     \
        name& operator=(name&&) = default;                                                                          \
        name& operator=(float val) {varname = val; return *this;}                                                   \
        virtual const std::string& getControlMode() const override                                                  \
        {                                                                                                           \
            return cmode;                                                                                           \
        }                                                                                                           \
        virtual void reset() override                                                                               \
        {                                                                                                           \
            varname = ControllerConstants::ValueNotSetNaN;                                                          \
        }                                                                                                           \
        virtual bool isValid() const override                                                                       \
        {                                                                                                           \
            return std::isfinite(varname);                                                                          \
        }                                                                                                           \
        DETAIL_ControlTargetBase_DEFAULT_METHOD_IMPLEMENTATION                                                      \
        static ControlTargetInfo<name> GetClassMemberInfo()                                                         \
        {                                                                                                           \
            ControlTargetInfo<name> cti;                                                                            \
            cti.addMemberVariable(&name::varname, #varname);                                                        \
            return cti;                                                                                             \
        }                                                                                                           \
    }

    make_ControlTarget1DoFActuator(ControlTarget1DoFActuatorPosition, position, ControlModes::Position1DoF);
    make_ControlTarget1DoFActuator(ControlTarget1DoFActuatorVelocity, velocity, ControlModes::Velocity1DoF);
    make_ControlTarget1DoFActuator(ControlTarget1DoFActuatorTorque, torque, ControlModes::Torque1DoF);
    make_ControlTarget1DoFActuator(ControlTarget1DoFActuatorCurrent, current, ControlModes::Current1DoF);

    class ControlTarget1DoFActuatorTorqueVelocity: public ControlTarget1DoFActuatorVelocity
    {
    public:
        float maxTorque;

        const std::string& getControlMode() const override
        {
            return ControlModes::VelocityTorque;
        }
        void reset() override
        {
            ControlTarget1DoFActuatorVelocity::reset();
            maxTorque = -1.0f; // if < 0, default value for joint is to be used
        }
        bool isValid() const override
        {
            return ControlTarget1DoFActuatorVelocity::isValid() && std::isfinite(maxTorque);
        }
        static ControlTargetInfo<ControlTarget1DoFActuatorTorqueVelocity> GetClassMemberInfo()
        {
            ControlTargetInfo<ControlTarget1DoFActuatorTorqueVelocity> cti;
            cti.addBaseClass<ControlTarget1DoFActuatorVelocity>();
            cti.addMemberVariable(&ControlTarget1DoFActuatorTorqueVelocity::maxTorque, "maxTorque");
            return cti;
        }
        DETAIL_ControlTargetBase_DEFAULT_METHOD_IMPLEMENTATION
    };

    class ActiveImpedanceControlTarget: public ControlTargetBase
    {
    public:
        float position;
        float kp;
        float kd;

        const std::string& getControlMode() const override
        {
            return ControlModes::ActiveImpedance;
        }
        void reset() override
        {
            position = 0;
            kp = 0;
            kd = 0;
        }
        bool isValid() const override
        {
            return std::isfinite(position) && kp >= 0;
        }
        static ControlTargetInfo<ActiveImpedanceControlTarget> GetClassMemberInfo()
        {
            ControlTargetInfo<ActiveImpedanceControlTarget> cti;
            cti.addMemberVariable(&ActiveImpedanceControlTarget::position, "position");
            cti.addMemberVariable(&ActiveImpedanceControlTarget::kp, "kp");
            cti.addMemberVariable(&ActiveImpedanceControlTarget::kd, "kd");
            return cti;
        }
        DETAIL_ControlTargetBase_DEFAULT_METHOD_IMPLEMENTATION
    };

#undef make_ControlTarget1DoFActuator
}

